import pytest

from curifactory import (
    CachersMismatchError,
    EmptyCachersError,
    InputSignatureError,
    OutputSignatureError,
    Record,
    aggregate,
    stage,
)
from curifactory.caching import Lazy, PickleCacher

# --------------------------
# @stage tests
# --------------------------


def test_stores_output_in_record(configured_test_manager):
    """Returned output from stage function should exist in record state."""

    @stage([], ["tester"])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    output_stage(record)
    assert record.state["tester"] == "hello world"


def test_stage_outputs_none(configured_test_manager):
    """Stage outputs set to None should be equivalent to [] and should not crash."""

    @stage([], outputs=None)
    def output_stage(record):
        x = 5
        del x

    record = Record(configured_test_manager, None)
    output_stage(record)


def test_stage_inputs_none(configured_test_manager):
    """Stage inputs set to None should be equivalent to [] and should not crash."""

    @stage(inputs=None, outputs=["tester"])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    output_stage(record)
    assert record.state["tester"] == "hello world"


def test_stores_multiple_outputs_in_record(configured_test_manager):
    """Multiple returned outputs from stage function should exist in record state."""

    @stage([], ["test1", "test2"])
    def output_stage(record):
        return "hello world", 13

    record = Record(configured_test_manager, None)
    output_stage(record)
    assert record.state["test1"] == "hello world"
    assert record.state["test2"] == 13


def test_returns_less_than_expected_errors(configured_test_manager):
    """A function that doesn't return the same number of objects as specified in the stage outputs should throw an OutputSignatureError."""

    @stage([], ["test1", "test2"])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        output_stage(record)


def test_empty_cachers_array_errors(configured_test_manager):
    """Don't let the user write [] for cachers list, as it always shortcircuits."""

    # NOTE: technically this could probably just be a special case handled by staging.py, but I think specifying [] for cachers is kind of unclear anyway.
    @stage([], [], [])
    def stage_that_does_nothing(record):
        i_am_very_important = 5 + 3
        del i_am_very_important

    record = Record(configured_test_manager, None)
    with pytest.raises(EmptyCachersError):
        stage_that_does_nothing(record)


def test_cachers_count_mismatch_errors(configured_test_manager):
    """Specifying a different number of cachers than output args should throw a CachersMismatchError."""

    @stage([], ["thing1", "thing2"], [PickleCacher])
    def bad_caching_vibes(record):
        return 1, 2

    record = Record(configured_test_manager, None)
    with pytest.raises(CachersMismatchError):
        bad_caching_vibes(record)


def test_cachers_correct_count(configured_test_manager):
    """Specifying the correct number of cachers for outputs works fine."""

    @stage([], ["thing1", "thing2"], [PickleCacher] * 2)
    def good_caching_vibes(record):
        return 1, 2

    record = Record(configured_test_manager, None)
    good_caching_vibes(record)
    # NOTE: will test caches correctly in caching tests


def test_return_tuple_for_one_output_errors(configured_test_manager):
    """Returning something that is a tuple does not work for a single specified output, this should throw an OutputSignatureError."""

    @stage([], ["my_output"])
    def return_tuple(record):
        multiple_values = (1, 2)
        return multiple_values

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        return_tuple(record)


def test_return_tuplefied_tuple_for_one_output(configured_test_manager):
    """Returning something that is a tuple as a single output value should be tuplefied again."""

    @stage([], ["my_output"])
    def return_tuple(record):
        multiple_values = (1, 2)
        return (multiple_values,)

    record = Record(configured_test_manager, None)
    return_tuple(record)
    assert record.state["my_output"] == (1, 2)


def test_return_array_for_one_output(configured_test_manager):
    """Returning an array as a single output should not have the same issues as a tuple, and should work fine."""

    @stage([], ["my_output"])
    def return_array(record):
        multiple_values = [1, 2]
        return multiple_values

    record = Record(configured_test_manager, None)
    return_array(record)
    assert record.state["my_output"] == [1, 2]


def test_takes_input_from_record(configured_test_manager):
    """A specified input in the stage should pass in that value from the pre-existing record state."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    record.state["my_input"] = 13
    replicate_input(record)
    assert record.state["replicated_input"] == 13


def test_mssing_input_errors(configured_test_manager):
    """A stage input that is not found in the record's state should throw a KeyError."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    with pytest.raises(KeyError):
        replicate_input(record)


def test_missing_input_with_suppress_and_no_default_errors(configured_test_manager):
    """If suppress_missing_inputs is set on the stage but no defaults were specified for the missing ones, we should get an InputSignatureError."""

    @stage(["my_input"], ["replicated_input"], suppress_missing_inputs=True)
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    with pytest.raises(InputSignatureError):
        replicate_input(record)


def test_input_overwritten_by_kwarg(configured_test_manager):
    """A directly passed input argument on the stage function should override whatever was in the record's state previously."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    record.state["my_input"] = 13
    replicate_input(record, my_input=15)
    assert record.state["replicated_input"] == 15


def test_missing_input_but_default_without_suppress_errors(configured_test_manager):
    """Even if there's a default value for an input argument, if suppress isn't specified and we don't pass a value in on the stage call, we should get an error."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input=15):
        return my_input

    record = Record(configured_test_manager, None)
    with pytest.raises(KeyError):
        replicate_input(record)


def test_missing_input_but_default_with_suppress(configured_test_manager):
    """If a stage has suppressed missing values and we don't pass the missing one in on the stage call, allow the default value to go through without error."""

    @stage(["my_input"], ["replicated_input"], suppress_missing_inputs=True)
    def replicate_input(record, my_input=15):
        return my_input

    record = Record(configured_test_manager, None)
    replicate_input(record)
    assert record.state["replicated_input"] == 15


def test_missing_input_but_kwarg_passed_with_suppress(configured_test_manager):
    """If we specify an input value directly in the stage call, allow it to go through even if it's missing from the record."""

    @stage(["my_input"], ["replicated_input"], suppress_missing_inputs=True)
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    replicate_input(record, my_input=15)
    assert record.state["replicated_input"] == 15


def test_missing_input_but_kwarg_passed_without_suppress(configured_test_manager):
    """If we specify an input value directly in the stage call, allow it to go through even if it's missing from the record and we don't have suppress enabled."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_input):
        return my_input

    record = Record(configured_test_manager, None)
    replicate_input(record, my_input=15)
    assert record.state["replicated_input"] == 15


def test_input_name_incorrect(configured_test_manager):
    """Stage input string names that do not match the function signature variable names should throw an InputSignatureError."""

    @stage(["my_input"], ["replicated_input"])
    def replicate_input(record, my_wrong_input):
        return my_wrong_input

    record = Record(configured_test_manager, None)
    record.state["my_input"] = 13
    with pytest.raises(InputSignatureError):
        replicate_input(record)


def test_lazy_obj_in_record(configured_test_manager):
    """Ensure a lazy object is put into state from a lazy output stage, and ensure that it reloads correctly when used."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    @stage(["tester"], outputs=["output"])
    def use_lazy(record, tester):
        return tester

    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == Lazy
    record.state.resolve = True

    use_lazy(record)
    assert record.state["output"] == "hello world"


def test_lazy_disabled_on_ignore_lazy(configured_test_manager):
    """Lazy objects should be replaced with str counterparts if ignore_lazy is set on the manager."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    configured_test_manager.ignore_lazy = True
    record = Record(configured_test_manager, None)
    output_stage(record)
    assert type(record.state["tester"]) == str
    record.state.resolve = False
    assert record.state["tester"] == "hello world"


def test_lazy_forced_on_manager_lazy(configured_test_manager):
    """All objects should be lazy if lazy is set on the manager."""

    @stage([], ["tester"], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    configured_test_manager.lazy = True
    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == Lazy
    record.state.resolve = True
    assert record.state["tester"] == "hello world"


def test_error_on_missing_lazy_cacher(configured_test_manager):
    """If no cacher is given for a lazy object, an error should be thrown."""

    @stage([], [Lazy("tester")])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        output_stage(record)
    # record.state.resolve = False
    # assert type(record.state["tester"]) == Lazy
    # assert type(record.state["tester"].cacher) == PickleCacher


def test_cacher_injected_on_manager_lazy(configured_test_manager):
    """If lazy is set on the manager and an object has no cacher, inject a PickleCacher."""

    @stage([], ["tester"])
    def output_stage(record):
        return "hello world"

    configured_test_manager.lazy = True
    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == Lazy
    assert type(record.state["tester"].cacher) == PickleCacher


def test_cacher_not_injected_on_manager_ignore_lazy(configured_test_manager):
    """If ignore_lazy is set on the manager and a Lazy object has no cacher, do not error."""
    # NOTE: in theory this behavior is so that if someone gives you buggy stages, you can "fix" the stage without having to actually go in and modify the cachers.

    @stage([], [Lazy("tester")])
    def output_stage(record):
        return "hello world"

    configured_test_manager.ignore_lazy = True
    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == str
    assert record.state["tester"] == "hello world"


def test_lazy_with_false_resolve_doesnot_resolve(configured_test_manager):
    """When a lazy object is set to not resolve, both directly accessing state, and
    a stage that utilizes that object should get the lazy object, not the underlying one.
    """

    @stage(None, [Lazy("tester", resolve=False)], [PickleCacher])
    def output_stage(record):
        return "hello world"

    @stage(["tester"])
    def use_lazy_stage(record, tester):
        assert type(tester) == Lazy
        assert type(tester.cacher) == PickleCacher
        assert tester.load() == "hello world"
        assert (
            tester.cacher.get_path()
            == "test/examples/data/cache/test_None_output_stage_tester.pkl"
        )

    record = Record(configured_test_manager, None)
    use_lazy_stage(output_stage(record))


def test_lazy_with_false_resolve_state_doesnot_resolve(configured_test_manager):
    """When a lazy object is set to not resolve, directly accessing state
    should return the lazy object, not the underlying one."""

    @stage(None, [Lazy("tester", resolve=False)], [PickleCacher])
    def output_stage(record):
        return "hello world"

    @stage(["tester"])
    def use_lazy_stage(record, tester):
        assert type(record.state["tester"]) == Lazy
        assert type(record.state["tester"].cacher) == PickleCacher
        assert record.state["tester"].load() == "hello world"

    record = Record(configured_test_manager, None)
    use_lazy_stage(output_stage(record))


def test_lazy_with_false_resolve_storefull_correct_cacher_path(configured_test_manager):
    """A lazy object set to not resolve in a store full manager should have a cacher
    whose path is still inside the regular cache path, not the run folder"""
    # TODO
    configured_test_manager.store_entire_run = True

    @stage(None, [Lazy("tester", resolve=False)], [PickleCacher])
    def output_stage(record):
        return "hello world"

    @stage(["tester"])
    def use_lazy_stage(record, tester):
        assert (
            tester.cacher.get_path()
            == "test/examples/data/cache/test_None_output_stage_tester.pkl"
        )

    record = Record(configured_test_manager, None)
    use_lazy_stage(output_stage(record))


# --------------------------
# @aggregate tests
# --------------------------


def test_aggregate_stores_output_in_record(configured_test_manager):
    """An aggregate output should exist in the record state."""

    @aggregate(None, ["output"])
    def small_aggregate(record, records):
        return "hello world"

    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])  # TODO: blank records array crashes??
    assert record.state["output"] == "hello world"


def test_aggregate_stores_multiple_outputs_in_record(configured_test_manager):
    """Multiple returned outputs from the stage should exist in the record state."""

    @aggregate(None, ["thing1", "thing2"])
    def small_aggregate(record, records):
        return "hello world", 13

    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    assert record.state["thing1"] == "hello world"
    assert record.state["thing2"] == 13


def test_aggregate_returns_less_than_expected_errors(configured_test_manager):
    """A function that doesn't return the same number of objects as specified in the aggregate outputs should throw an OutputSignatureError."""

    @aggregate(None, ["test1", "test2"])
    def output_stage(record, records):
        return "hello world"

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        output_stage(record, [record])


def test_aggregate_empty_cachers_array_errors(configured_test_manager):
    """Don't let the user write [] for cachers list, as it always shortcircuits."""

    # NOTE: technically this could probably just be a special case handled by staging.py, but I think specifying [] for cachers is kind of unclear anyway.
    @aggregate([], [], [])
    def aggregate_stage_that_does_nothing(record, records):
        i_am_very_important = 5 + 3
        del i_am_very_important

    record = Record(configured_test_manager, None)
    with pytest.raises(EmptyCachersError):
        aggregate_stage_that_does_nothing(record, [record])


def test_aggregate_cachers_count_mismatch_errors(configured_test_manager):
    """Specifying a different number of cachers than output args should throw a CachersMismatchError."""

    @aggregate(None, ["thing1", "thing2"], [PickleCacher])
    def bad_caching_vibes(record, records):
        return 1, 2

    record = Record(configured_test_manager, None)
    with pytest.raises(CachersMismatchError):
        bad_caching_vibes(record, [record])


def test_aggregate_cachers_correct_count(configured_test_manager):
    """Specifying the correct number of cachers for outputs works fine."""

    @aggregate(None, ["thing1", "thing2"], [PickleCacher] * 2)
    def good_caching_vibes(record, records):
        return 1, 2

    record = Record(configured_test_manager, None)
    good_caching_vibes(record, [record])
    # NOTE: will test caches correctly in caching tests


def test_aggregate_return_tuple_for_one_output_errors(configured_test_manager):
    """Returning something that is a tuple does not work for a single specified output, this should throw an OutputSignatureError."""

    @aggregate(None, ["my_output"])
    def return_tuple(record, records):
        multiple_values = (1, 2)
        return multiple_values

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        return_tuple(record, [record])


def test_aggregate_return_tuplefied_tuple_for_one_output(configured_test_manager):
    """Returning something that is a tuple as a single output value should be tuplefied again."""

    @aggregate(None, ["my_output"])
    def return_tuple(record, records):
        multiple_values = (1, 2)
        return (multiple_values,)

    record = Record(configured_test_manager, None)
    return_tuple(record, [record])
    assert record.state["my_output"] == (1, 2)


def test_aggregate_return_array_for_one_output(configured_test_manager):
    """Returning an array as a single output should not have the same issues as a tuple, and should work fine."""

    @aggregate(None, ["my_output"])
    def return_array(record, records):
        multiple_values = [1, 2]
        return multiple_values

    record = Record(configured_test_manager, None)
    return_array(record, [record])
    assert record.state["my_output"] == [1, 2]


def test_aggregate_gets_records_states(configured_test_manager):
    """Inside the aggregate, we should be able to get values from the states of the manually passed in records."""

    @aggregate(None, ["output"])
    def small_aggregate(record, records):
        total = 0
        for r in records:
            total += r.state["input"]
        return total

    record1 = Record(configured_test_manager, None)
    record2 = Record(configured_test_manager, None)
    record3 = Record(configured_test_manager, None)

    record1.state["input"] = 3
    record2.state["input"] = 2
    small_aggregate(record3, [record1, record2])
    assert record3.state["output"] == 5


def test_aggregate_gets_no_manager_records(configured_test_manager):
    """Calling an aggregate stage without records and with no records on the manager should not crash."""

    @aggregate(None, ["output"])
    def small_aggregate(record, records):
        total = 0
        for r in records:
            total += r.state["input"]
        return total

    record = Record(configured_test_manager, None, hide=True)
    small_aggregate(record)
    assert record.state["output"] == 0


def test_aggregate_gets_manager_records_states(configured_test_manager):
    """Inside the aggregate, we should be passed all records on the manager by default, if no records explicitly passed."""

    @aggregate(None, ["output"])
    def small_aggregate(record, records):
        total = 0
        for r in records:
            total += r.state["input"]
        return total

    record1 = Record(configured_test_manager, None)
    record2 = Record(configured_test_manager, None)
    record3 = Record(configured_test_manager, None, hide=True)

    record1.state["input"] = 3
    record2.state["input"] = 2

    small_aggregate(record3)
    assert record3.state["output"] == 5


def test_aggregate_gets_manager_records_except_own(configured_test_manager):
    """A record with an aggregate stage and no explicitly provided records should grab all records on the manager
    minus that currently executing record."""

    @aggregate(None, ["output"])
    def small_aggregate(record, records):
        total = 0
        for r in records:
            total += r.state["input"]
        return total

    record1 = Record(configured_test_manager, None)
    record2 = Record(configured_test_manager, None)
    record3 = Record(configured_test_manager, None)

    record1.state["input"] = 3
    record2.state["input"] = 2
    record3.state["input"] = 1  # note that this is for the test and
    # doesn't actually make sense - an aggregate stage should always
    # be the first stage of the record.

    small_aggregate(record3)
    assert record3.state["output"] == 5


def test_aggregate_lazy_obj_in_record(configured_test_manager):
    """Ensure a lazy object is put into state from a lazy output aggregate stage, and ensure that it reloads correctly when used."""

    @aggregate(None, [Lazy("output")], cachers=[PickleCacher])
    def small_aggregate(record, records):
        total = 0
        for r in records:
            total += r.state["input"]
        return total

    @stage(["output"], outputs=["actual_output"])
    def use_lazy(record, output):
        return output

    record1 = Record(configured_test_manager, None)
    record2 = Record(configured_test_manager, None)
    record3 = Record(configured_test_manager, None, hide=True)

    record1.state["input"] = 3
    record2.state["input"] = 2

    small_aggregate(record3)
    record3.state.resolve = False
    assert type(record3.state["output"]) == Lazy
    record3.state.resolve = True

    use_lazy(record3)
    assert record3.state["actual_output"] == 5


def test_aggregate_lazy_disabled_on_ignore_lazy(configured_test_manager):
    """Lazy objects should be replaced with str counterparts if ignore_lazy is set on the manager."""

    @aggregate(None, [Lazy("output")], cachers=[PickleCacher])
    def small_aggregate(record, records):
        return "hello world"

    configured_test_manager.ignore_lazy = True
    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    record.state.resolve = False
    assert type(record.state["output"]) == str
    assert record.state["output"] == "hello world"


def test_aggregate_lazy_forced_on_manager_lazy(configured_test_manager):
    """All objects should be lazy if lazy is set on the manager."""

    @aggregate(None, ["output"], cachers=[PickleCacher])
    def small_aggregate(record, records):
        return "hello world"

    configured_test_manager.lazy = True
    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    record.state.resolve = False
    assert type(record.state["output"]) == Lazy
    record.state.resolve = True
    assert record.state["output"] == "hello world"


def test_aggregate_error_on_missing_lazy_cacher(configured_test_manager):
    """If no cacher is given for a lazy object, an error should be thrown."""

    @aggregate(None, [Lazy("output")])
    def small_aggregate(record, records):
        return "hello world"

    record = Record(configured_test_manager, None)
    with pytest.raises(OutputSignatureError):
        small_aggregate(record, [record])


def test_aggregate_cacher_injected_on_manager_lazy(configured_test_manager):
    """If lazy is set on the manager and an object has no cacher, inject a PickleCacher."""

    @aggregate(None, ["output"])
    def small_aggregate(record, records):
        return "hello world"

    configured_test_manager.lazy = True
    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    record.state.resolve = False
    assert type(record.state["output"]) == Lazy
    assert type(record.state["output"].cacher) == PickleCacher


def test_aggregate_cacher_not_injected_on_manager_ignore_lazy(configured_test_manager):
    """If ignore_lazy is set on the manager and a Lazy object has no cacher, do not error."""
    # NOTE: in theory this behavior is so that if someone gives you buggy stages, you can "fix" the stage without having to actually go in and modify the cachers.

    @aggregate(None, [Lazy("output")])
    def small_aggregate(record, records):
        return "hello world"

    configured_test_manager.ignore_lazy = True
    record = Record(configured_test_manager, None)
    small_aggregate(record, [record])
    record.state.resolve = False
    assert type(record.state["output"]) == str
    assert record.state["output"] == "hello world"


def test_aggregate_auto_resolve_lazy_state(configured_test_manager):
    """A lazy cached object in state should auto-resolve when we access it in an aggregate."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    @aggregate(None, ["reproduced_output"])
    def small_aggregate(record, records):
        return records[0].state["tester"]

    record1 = Record(configured_test_manager, None)
    record2 = Record(configured_test_manager, None, hide=True)

    output_stage(record1)
    record1.state.resolve = False
    assert type(record1.state["tester"]) == Lazy
    record1.state.resolve = True

    small_aggregate(record2, [record1])
    assert record2.state["reproduced_output"] == "hello world"


def test_aggregate_populates_inputs_when_all_records_have_artifact(
    configured_test_manager,
):
    """Inputs for an aggregate state should correctly populate the input-specific argument when all records have
    the requested artifact."""

    @stage([], ["value"])
    def make_value(record):
        return 3

    @aggregate(["value"], ["total"])
    def combine_values(record, records, value):
        total = 0
        assert len(value) == len(records)
        for r in records:
            assert r in list(value.keys())
        for r, val in value.items():
            total += val

        return total

    r0 = Record(configured_test_manager, None)
    r1 = Record(configured_test_manager, None)
    r2 = Record(configured_test_manager, None)

    combine_values(r2, [make_value(r0), make_value(r1)])
    assert r2.state["total"] == 6


def test_aggregate_populates_inputs_when_some_records_have_artifact(
    configured_test_manager,
):
    """Inputs for an aggregate state should correctly populate the input-specific argument when only some of
    the records have the requested artifact."""

    @stage([], ["value"])
    def make_value(record):
        return 3

    @stage([], ["not_value"])
    def make_not_value(record):
        return 3

    @aggregate(["value", "not_value"], ["total"])
    def combine_values(record, records, value, not_value):
        total = 0
        assert len(value) == 1
        assert len(not_value) == 1
        assert list(value.keys())[0] in records
        assert list(not_value.keys())[0] in records
        for r in records:
            if r in value:
                total += value[r]
            elif r in not_value:
                total += 2 * not_value[r]

        return total

    r0 = Record(configured_test_manager, None)
    r1 = Record(configured_test_manager, None)
    r2 = Record(configured_test_manager, None)

    combine_values(r2, [make_value(r0), make_not_value(r1)])
    assert r2.state["total"] == 9


def test_aggregate_populates_inputs_when_no_records_have_artifact(
    configured_test_manager,
):
    """Inputs for an aggregate state should correctly populate the input-specific argument when none of the
    records have the requested artifact."""

    @stage([], ["not_value"])
    def make_not_value(record):
        return 3

    @aggregate(["value"], ["total"])
    def combine_values(record, records, value):
        total = 0
        assert len(value) == 0
        assert len(records) != 0

        for r, val in value.items():
            total += val

        return total

    r0 = Record(configured_test_manager, None)
    r1 = Record(configured_test_manager, None)
    r2 = Record(configured_test_manager, None)

    combine_values(r2, [make_not_value(r0), make_not_value(r1)])
    assert r2.state["total"] == 0


def test_aggregate_can_cross_reference_inputs_by_record(configured_test_manager):
    """In a case where two input records to an aggregate state have one common input in state and each have
    one not in the other, you should be able to correctly key the dictionaries based on the record.
    """

    @stage([], ["value", "common"])
    def make_value(record):
        return 3, 1

    @stage([], ["not_value", "common"])
    def make_not_value(record):
        return 8, 2

    @aggregate(["value", "not_value", "common"], ["total"])
    def combine_values(record, records, value, not_value, common):
        total = 0

        assert len(value) == 1
        assert len(not_value) == 1
        assert len(common) == len(records)
        assert list(value.keys())[0] in records
        assert list(not_value.keys())[0] in records
        for r in records:
            assert r in common.keys()
        for r in value:
            assert r in common.keys()
        for r in not_value:
            assert r in common.keys()

        for r, val in value.items():
            total += val
            total += common[r]

        for r, val in not_value.items():
            total += val
            total += common[r]

        return total

    r0 = Record(configured_test_manager, None)
    r1 = Record(configured_test_manager, None)
    r2 = Record(configured_test_manager, None)

    combine_values(r2, [make_value(r0), make_not_value(r1)])
    assert r2.state["total"] == 14


def test_aggregate_populates_inputs_when_all_records_have_lazy_artifact(
    configured_test_manager,
):
    """Inputs for an aggregate state should correctly populate the input-specific argument when all records have
    the requested artifact, and that it resolves correctly if lazy."""

    @stage([], [Lazy("value")], [PickleCacher])
    def make_value(record):
        return 3

    @aggregate(["value"], ["total"])
    def combine_values(record, records, value):
        total = 0
        assert len(value) == len(records)
        for r in records:
            assert r in list(value.keys())
        for r, val in value.items():
            total += val

        return total

    r0 = Record(configured_test_manager, None)
    r1 = Record(configured_test_manager, None)
    r2 = Record(configured_test_manager, None)

    combine_values(r2, [make_value(r0), make_value(r1)])
    assert r2.state["total"] == 6



import pytest

import curifactory as cf
from curifactory.caching import PickleCacher


def test_child_records_empty_for_blank_records(configured_test_manager):
    """Records with no relation to eachother should have empty child_records"""
    configured_test_manager.map_mode = True
    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test0"))
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters("test1"))

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    r1 = dag.records[1]
    assert len(dag.child_records(r0)) == 0
    assert len(dag.child_records(r1)) == 0


def test_child_records_after_make_copy(configured_test_manager):
    """A record that was copied should show the copied record as one of its children."""
    configured_test_manager.map_mode = True

    @cf.stage()
    def do_thing(record):
        t = 5 + 2
        print(t)

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test0"))
    r0 = do_thing(r0)
    r1 = r0.make_copy(cf.ExperimentParameters("test1"))

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    r1 = dag.records[1]
    r0_children = dag.child_records(r0)
    assert len(r0_children) == 1
    assert len(dag.child_records(r1)) == 0
    assert r0_children[0] == r1


def test_output_used_check_when_no_following_stages(configured_test_manager):
    """is_output_used_anywhere should return false when there are no following stages or
    records."""
    configured_test_manager.map_mode = True

    @cf.stage(outputs=["thing"])
    def thing1(record):
        return "green eggs and ham"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test0"))
    r0 = thing1(r0)

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    assert not dag.is_output_used_anywhere(r0, 1, "thing")


def test_output_used_check_when_record_copied_butunused(configured_test_manager):
    """is_output_used_anywhere should return false when there are no following stages in subsequent
    records that use it."""
    configured_test_manager.map_mode = True

    @cf.stage(outputs=["thing"])
    def thing1(record):
        return "green eggs and ham"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test0"))
    r0 = thing1(r0)

    r0.make_copy(cf.ExperimentParameters("test1"))

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    assert not dag.is_output_used_anywhere(r0, 1, "thing")


def test_output_used_check_in_aggregate_when_not_expected(configured_test_manager):
    """is_output_used_anywhere should return false when there is an aggregate using this record
    but does not list the variable as part of expected state."""
    configured_test_manager.map_mode = True

    @cf.stage(outputs=["thing"])
    def thing1(record):
        return "green eggs and ham"

    @cf.aggregate(outputs=["things"])
    def all_the_things(record, records):
        return "no"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test0"))
    r0 = thing1(r0)

    r1 = cf.Record(configured_test_manager, None)
    r1 = all_the_things(r1, [r0])

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    assert not dag.is_output_used_anywhere(r0, 1, "thing")


def test_output_used_check_in_aggregate(configured_test_manager):
    """is_output_used_anywhere should return true when there is an aggregate using this record."""
    configured_test_manager.map_mode = True

    @cf.stage(outputs=["thing"])
    def thing1(record):
        return "green eggs and ham"

    @cf.aggregate(inputs=["thing"], outputs=["things"])
    def all_the_things(record, records, thing):
        return "no"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test0"))
    r0 = thing1(r0)

    r1 = cf.Record(configured_test_manager, None)
    r1 = all_the_things(r1, [r0])

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    assert dag.is_output_used_anywhere(r0, 1, "thing")


def test_output_used_check_finds_output_in_later_input(configured_test_manager):
    """An output that's actually used later should be found."""
    configured_test_manager.map_mode = True

    @cf.stage(outputs=["thing"])
    def thing1(record):
        return "green eggs and ham"

    @cf.stage(inputs=["thing"], outputs=["another_thing"])
    def thing2(record, thing):
        return "Sam I am"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test0"))
    thing1(r0)
    thing2(r0)
    # thing2(thing1(r0))

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    assert dag.is_output_used_anywhere(r0, 1, "thing")


def test_output_used_check_finds_output_in_later_input_of_copy(configured_test_manager):
    """An output that's used as an input in a later copy of a record should be found."""
    configured_test_manager.map_mode = True

    @cf.stage(outputs=["thing"])
    def thing1(record):
        return "green eggs and ham"

    @cf.stage(inputs=["thing"], outputs=["another_thing"])
    def thing2(record, thing):
        return "Sam I am"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test0"))
    thing1(r0)
    r1 = r0.make_copy(cf.ExperimentParameters("test1"))
    thing2(r1)

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    assert dag.is_output_used_anywhere(r0, 1, "thing")


def test_single_stage_is_leaf(configured_test_manager):
    """Make sure a single stage is in fact a leaf."""
    configured_test_manager.map_mode = True

    @cf.stage()
    def void_stares_back(record):
        print("If you stare into the abyss")

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    void_stares_back(r0)

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    assert dag.is_leaf(r0, "void_stares_back")


def test_output_used_elsewhere_is_not_leaf(configured_test_manager):
    configured_test_manager.map_mode = True

    @cf.stage(outputs=["thing"])
    def thing1(record):
        return "green eggs and ham"

    @cf.stage(inputs=["thing"], outputs=["another_thing"])
    def thing2(record, thing):
        return "Sam I am"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test0"))
    thing2(thing1(r0))

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    r0 = dag.records[0]
    assert not dag.is_leaf(r0, "thing1")
    assert dag.is_leaf(r0, "thing2")


@pytest.mark.parametrize(
    "outputs,kwargs,suppress_missing,expect_error",
    [
        (None, {}, False, True),
        (None, {"nothing": 1}, False, False),
        (None, {"nothing": 1}, True, False),
        (None, {}, True, False),
        # having an output shouldn't actually change the handling
        (["thing"], {}, False, True),
        (["thing"], {"nothing": 1}, False, False),
        (["thing"], {"nothing": 1}, True, False),
        (["thing"], {}, True, False),
    ],
)
def test_stage_with_missing_inputs_handled_correctly(
    configured_test_manager, outputs, kwargs, suppress_missing, expect_error
):
    """A stage requesting an input that doesn't exist in state should appropriately
    handle kwargs and suppress_missing passed to a stage in determining whether
    an error gets thrown or not."""
    configured_test_manager.map_mode = True

    @cf.stage(
        inputs=["nothing"], outputs=outputs, suppress_missing_inputs=suppress_missing
    )
    def do_nothing(record, nothing):
        return 6

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_nothing(r0, **kwargs)
    configured_test_manager.map_mode = False

    # TODO: check for error her
    if expect_error:
        with pytest.raises(KeyError):
            configured_test_manager.map_records()
            dag = configured_test_manager.map
            dag.determine_execution_list()
    else:
        configured_test_manager.map_records()
        dag = configured_test_manager.map
        dag.determine_execution_list()

        # expect to always execute a stage that has missing inputs
        assert dag.execution_list == [(0, "do_nothing")]


def test_single_record_execution_list_with_no_outputs(configured_test_manager):
    """A stage with no outputs should count as a leaf and should be in the execution list."""
    configured_test_manager.map_mode = True

    @cf.stage()
    def do_nothing(record):
        thing = 5 + 1
        thing += 1

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_nothing(r0)
    configured_test_manager.map_mode = False
    configured_test_manager.map_records()

    dag = configured_test_manager.map
    dag.determine_execution_list()
    assert dag.execution_list == [(0, "do_nothing")]


@pytest.mark.parametrize(
    "cached,overwrite_stages,expected_execution_list",
    [
        (False, [], [(0, "do_thing"), (0, "do_nothing")]),
        (True, [], [(0, "do_nothing")]),
        (False, ["do_thing"], [(0, "do_thing"), (0, "do_nothing")]),
        (True, ["do_thing"], [(0, "do_thing"), (0, "do_nothing")]),
        (False, ["do_thing", "do_nothing"], [(0, "do_thing"), (0, "do_nothing")]),
        (True, ["do_thing", "do_nothing"], [(0, "do_thing"), (0, "do_nothing")]),
        (True, ["do_nothing"], [(0, "do_nothing")]),
    ],
)
def test_single_record_execution_list_with_no_outputs_and_inputs(
    configured_test_manager, cached, overwrite_stages, expected_execution_list
):
    """A stage with no outputs (but does have inputs) should count as a leaf and should
    be in the execution list."""
    configured_test_manager.map_mode = True
    configured_test_manager.overwrite_stages = overwrite_stages

    @cf.stage(outputs=["thing"])
    def do_thing(record):
        return 5

    @cf.stage(inputs=["thing"])
    def do_nothing(record, thing):
        thing = 5 + 1
        thing += 1

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_nothing(do_thing(r0))
    configured_test_manager.map_mode = False
    configured_test_manager.map_records()

    dag = configured_test_manager.map

    dag.artifacts[0].cached = cached
    dag.determine_execution_list()
    assert dag.execution_list == expected_execution_list


@pytest.mark.parametrize(
    "cached,overwrite_stages,expected_execution_list",
    [
        (False, [], [(0, "thing1")]),
        (True, [], []),
        (False, ["thing1"], [(0, "thing1")]),
        (True, ["thing1"], [(0, "thing1")]),
    ],
)
def test_single_record_execution_lists(
    configured_test_manager, cached, overwrite_stages, expected_execution_list
):
    """DAG-based execution determination should be correct for all possible
    cached/overwrite combinations for one stage in one record:

        (r0:thing1)
    """
    configured_test_manager.map_mode = True
    configured_test_manager.overwrite_stages = overwrite_stages

    @cf.stage(outputs=["thing1"])
    def thing1(record):
        return 1

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    thing1(r0)

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    dag.artifacts[0].cached = cached
    dag.determine_execution_list()

    assert dag.execution_list == expected_execution_list


@pytest.mark.parametrize(
    "cached,overwrite_stages,expected_execution_list",
    [
        # fmt: off
        ([False, False], [], [(0, "thing1"), (0, "thing2")]),
        ([True, False], [], [(0, "thing2")]),
        ([True, True], [], []),
        ([False, True], [], []),
        ([False, True], ["thing2"], [(0, "thing1"), (0, "thing2")]),
        ([True, False], ["thing2"], [(0, "thing2")]),
        ([True, True], ["thing2"], [(0, "thing2")]),
        ([True, True], ["thing1"], [(0, "thing1"), (0, "thing2")]),
        ([False, True], ["thing1"], [(0, "thing1"), (0, "thing2")]),
        ([False, False], ["thing1"], [(0, "thing1"), (0, "thing2")]),
        # fmt: on
    ],
)
def test_single_record_double_stage_execution_lists(
    configured_test_manager, cached, overwrite_stages, expected_execution_list
):
    """DAG-based execution determination should be correct for all possible
    cached/overwrite combinations for two stages in one record:

        (r0:thing1)--(r0:thing2)
    """
    configured_test_manager.map_mode = True
    configured_test_manager.overwrite_stages = overwrite_stages

    @cf.stage(outputs=["thing1"])
    def thing1(record):
        return 1

    @cf.stage(inputs=["thing1"], outputs=["thing2"])
    def thing2(record, thing1):
        return thing1 + 1

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    thing2(thing1(r0))

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    dag.artifacts[0].cached = cached[0]
    dag.artifacts[1].cached = cached[1]
    dag.determine_execution_list()

    assert len(dag.execution_list) == len(expected_execution_list)
    assert set(dag.execution_list) == set(expected_execution_list)


@pytest.mark.parametrize(
    "cached,overwrite_stages,expected_execution_list",
    [
        # fmt: off
        ([False, False, False], [], [(0, "thing1"), (0, "thing2"), (0, "thing3")]),
        ([False, False, True], [], []),
        ([False, True, False], [], [(0, "thing3")]),
        ([True, False, False], [], [(0, "thing2"), (0, "thing3")]),
        ([True, True, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (0, "thing3")]),
        ([True, False, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (0, "thing3")]),
        ([True, False, True], ["thing2"], [(0, "thing2"), (0, "thing3")]),
        ([False, False, True], ["thing2"], [(0, "thing1"), (0, "thing2"), (0, "thing3")]),
        # fmt: on
    ],
)
def test_single_record_triple_stage_execution_lists(
    configured_test_manager, cached, overwrite_stages, expected_execution_list
):
    """DAG-based execution determination should be correct for all possible
    cached/overwrite combinations for three stages in one record:

        (r0:thing1)--(r0:thing2)--(r0:thing3)
    """
    configured_test_manager.map_mode = True
    configured_test_manager.overwrite_stages = overwrite_stages

    @cf.stage(outputs=["thing1"])
    def thing1(record):
        return 1

    @cf.stage(inputs=["thing1"], outputs=["thing2"])
    def thing2(record, thing1):
        return thing1 + 1

    @cf.stage(inputs=["thing2"], outputs=["thing3"])
    def thing3(record, thing2):
        return thing2 + 1

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    thing3(thing2(thing1(r0)))

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    dag.artifacts[0].cached = cached[0]
    dag.artifacts[1].cached = cached[1]
    dag.artifacts[2].cached = cached[2]
    dag.determine_execution_list()

    assert len(dag.execution_list) == len(expected_execution_list)
    assert set(dag.execution_list) == set(expected_execution_list)


@pytest.mark.parametrize(
    "cached,overwrite_stages,expected_execution_list",
    [
        # fmt: off
        ([False, False, False], [], [(0, "thing1"), (0, "thing2"), (0, "thing3")]),
        ([True, False, False], [], [(0, "thing2"), (0, "thing3")]),
        ([True, True, True], [], []),
        ([False, True, True], [], []),
        ([False, False, True], [], [(0, "thing1"), (0, "thing2")]),
        ([False, True, False], [], [(0, "thing1"), (0, "thing3")]),
        ([True, True, False], [], [(0, "thing3")]),
        ([True, False, True], [], [(0, "thing2")]),
        ([True, True, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (0, "thing3")]),
        ([False, True, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (0, "thing3")]),
        ([False, False, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (0, "thing3")]),
        ([True, True, True], ["thing2"], [(0, "thing2")]),
        ([True, True, True], ["thing3"], [(0, "thing3")]),
        ([False, True, True], ["thing2"], [(0, "thing1"), (0, "thing2")]),
        ([False, True, True], ["thing3"], [(0, "thing1"), (0, "thing3")]),
        # fmt: on
    ],
)
def test_single_record_triple_stage_nonsinglechain_execution_lists(
    configured_test_manager, cached, overwrite_stages, expected_execution_list
):
    """DAG-based execution determination should be correct for all possible
    cached/overwrite combinations for three stages in one record, where two stages
    don't impact each other but both use the same outputs from the first stage

                    (r0:thing2)
                   /
        (r0:thing1)
                   \
                    (r0:thing3)
    """
    configured_test_manager.map_mode = True
    configured_test_manager.overwrite_stages = overwrite_stages

    @cf.stage(outputs=["thing1"])
    def thing1(record):
        return 1

    @cf.stage(inputs=["thing1"], outputs=["thing2"])
    def thing2(record, thing1):
        return thing1 + 1

    @cf.stage(inputs=["thing1"], outputs=["thing3"])
    def thing3(record, thing1):
        return thing1 + 2

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    thing1(r0)
    thing2(r0)
    thing3(r0)

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    dag.artifacts[0].cached = cached[0]
    dag.artifacts[1].cached = cached[1]
    dag.artifacts[2].cached = cached[2]
    dag.determine_execution_list()

    assert len(dag.execution_list) == len(expected_execution_list)
    assert set(dag.execution_list) == set(expected_execution_list)


@pytest.mark.parametrize(
    "cached,overwrite_stages,expected_execution_list",
    [
        # fmt: off
        ([False, False, False], [], [(0, "thing1"), (0, "thing2"), (1, "thing3")]),
        ([True, False, False], [], [(0, "thing2"), (1, "thing3")]),
        ([True, True, True], [], []),
        ([False, True, True], [], []),
        ([False, False, True], [], [(0, "thing1"), (0, "thing2")]),
        ([False, True, False], [], [(0, "thing1"), (1, "thing3")]),
        ([True, True, False], [], [(1, "thing3")]),
        ([True, False, True], [], [(0, "thing2")]),
        ([True, True, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (1, "thing3")]),
        ([False, True, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (1, "thing3")]),
        ([False, False, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (1, "thing3")]),
        ([True, True, True], ["thing2"], [(0, "thing2")]),
        ([True, True, True], ["thing3"], [(1, "thing3")]),
        ([False, True, True], ["thing2"], [(0, "thing1"), (0, "thing2")]),
        ([False, True, True], ["thing3"], [(0, "thing1"), (1, "thing3")]),
        # fmt: on
    ],
)
def test_double_record_triple_stage_execution_lists(
    configured_test_manager, cached, overwrite_stages, expected_execution_list
):
    """DAG-based execution determination should be correct for all possible
    cached/overwrite combinations for three stages in two records, where one
    record is a copy of the other after the first stage:

                    (r0:thing2)
                   /
        (r0:thing1)
                   \
                    (r1:thing3)
    """
    configured_test_manager.map_mode = True
    configured_test_manager.overwrite_stages = overwrite_stages

    @cf.stage(outputs=["thing1"])
    def thing1(record):
        return 1

    @cf.stage(inputs=["thing1"], outputs=["thing2"])
    def thing2(record, thing1):
        return thing1 + 1

    @cf.stage(inputs=["thing1"], outputs=["thing3"])
    def thing3(record, thing1):
        return thing1 + 2

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    thing1(r0)
    r1 = r0.make_copy(cf.ExperimentParameters("test2"))
    thing2(r0)
    thing3(r1)

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    dag.artifacts[0].cached = cached[0]
    dag.artifacts[1].cached = cached[1]
    dag.artifacts[2].cached = cached[2]
    dag.determine_execution_list()

    assert len(dag.execution_list) == len(expected_execution_list)
    assert set(dag.execution_list) == set(expected_execution_list)


@pytest.mark.parametrize(
    "cached,overwrite_stages,expected_execution_list",
    [
        # fmt: off
        ([False, False, False, False], [], [(0, "thing1"), (0, "thing2"), (1, "thing3"), (2, "agg_things")]),
        ([False, False, False, True], [], []),
        ([False, False, True, False], [], [(0, "thing1"), (0, "thing2"), (2, "agg_things")]),
        ([False, True, False, False], [], [(0, "thing1"), (1, "thing3"), (2, "agg_things")]),
        ([False, True, True, False], [], [(2, "agg_things")]),
        ([True, True, True, False], [], [(2, "agg_things")]),
        ([True, True, True, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (1, "thing3"), (2, "agg_things")]),
        ([False, True, True, True], ["thing1"], [(0, "thing1"), (0, "thing2"), (1, "thing3"), (2, "agg_things")]),
        ([True, True, False, True], ["thing2"], [(0, "thing2"), (1, "thing3"), (2, "agg_things")]),
        ([True, False, True, True], ["thing3"], [(0, "thing2"), (1, "thing3"), (2, "agg_things")]),
        ([True, True, True, True], ["thing2"], [(0, "thing2"), (2, "agg_things")]),
        ([True, True, True, True], ["thing3"], [(1, "thing3"), (2, "agg_things")]),
        ([False, False, False, True], ["thing2"], [(0, "thing1"), (0, "thing2"), (1, "thing3"), (2, "agg_things")]),
        ([False, False, False, True], ["thing3"], [(0, "thing1"), (0, "thing2"), (1, "thing3"), (2, "agg_things")]),
        ([False, False, True, True], ["thing2"], [(0, "thing1"), (0, "thing2"), (2, "agg_things")]),
        ([False, True, False, True], ["thing3"], [(0, "thing1"), (1, "thing3"), (2, "agg_things")]),
        # fmt: on
    ],
)
def test_triple_record_quadruple_stage_execution_lists(
    configured_test_manager, cached, overwrite_stages, expected_execution_list
):
    """DAG-based execution determination should be correct for all possible
    cached/overwrite combinations for four stages in three records, where one
    record is a copy of the other after the first stage and the final stage
    aggregates the two former records:

                    (r0:thing2)
                   /           \
        (r0:thing1)             (r2:thing4)
                   \\           /
                    (r1:thing3)
    """
    configured_test_manager.map_mode = True
    configured_test_manager.overwrite_stages = overwrite_stages

    @cf.stage(outputs=["thing1"])
    def thing1(record):
        return 1

    @cf.stage(inputs=["thing1"], outputs=["thing"])
    def thing2(record, thing1):
        return thing1 + 1

    @cf.stage(inputs=["thing1"], outputs=["thing"])
    def thing3(record, thing1):
        return thing1 + 2

    @cf.aggregate(inputs=["thing"], outputs=["things"])
    def agg_things(record, records, thing):
        total = 0
        for r, value in thing.items():
            total += value
        return total

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    thing1(r0)
    r1 = r0.make_copy(cf.ExperimentParameters("test2"))
    thing2(r0)
    thing3(r1)
    r2 = cf.Record(configured_test_manager, None)
    agg_things(r2, [r0, r1])

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    dag.artifacts[0].cached = cached[0]
    dag.artifacts[1].cached = cached[1]
    dag.artifacts[2].cached = cached[2]
    dag.artifacts[3].cached = cached[3]
    dag.determine_execution_list()

    assert len(dag.execution_list) == len(expected_execution_list)
    assert set(dag.execution_list) == set(expected_execution_list)


# TODO: tests with two records two stages (single record first stage, two records run second stage)

# TODO: also tests where the outputs of a stage are each used as an input in a different stage


def test_dag_will_not_force_recompute_of_similar_stages(configured_test_manager):
    """A stage that is called multiple times with the same arguments and is not _initially_ cached
    should not be run more than once. (The fact that the DAG execution list doesn't update shouldn't
    preclude stage from determining a stage still doesn't actually need to run if it finds the right
    cached values.)"""

    run_count = 0

    @cf.stage(outputs=["thing"], cachers=[PickleCacher])
    def do_thing(record):
        nonlocal run_count
        run_count += 1
        return "hi"

    configured_test_manager.map_mode = True
    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_thing(r0)
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_thing(r1)

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    assert len(dag.execution_list) == 2
    assert run_count == 0

    actual_r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_thing(actual_r0)
    actual_r1 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_thing(actual_r1)

    assert run_count == 1


def test_dag_will_not_force_recompute_of_similar_aggregate_stages(
    configured_test_manager,
):
    """An agg stage that is called multiple times with the same arguments and is not _initially_ cached
    should not be run more than once. (The fact that the DAG execution list doesn't update shouldn't
    preclude stage from determining a stage still doesn't actually need to run if it finds the right
    cached values.)"""

    run_count = 0

    @cf.aggregate(outputs=["thing"], cachers=[PickleCacher])
    def do_thing(record, records):
        nonlocal run_count
        run_count += 1
        return "hi"

    configured_test_manager.map_mode = True
    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_thing(r0, [])
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_thing(r1, [])

    configured_test_manager.map_mode = False
    configured_test_manager.map_records()
    dag = configured_test_manager.map

    assert len(dag.execution_list) == 2
    assert run_count == 0

    actual_r0 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_thing(actual_r0, [])
    actual_r1 = cf.Record(configured_test_manager, cf.ExperimentParameters("test"))
    do_thing(actual_r1, [])

    assert run_count == 1



"""Testing record functionality."""

import json
import os

from pytest_mock import mocker  # noqa: F401 -- flake8 doesn't see it's used as fixture

from curifactory import ExperimentParameters, Record, aggregate, hashing, stage
from curifactory.caching import JsonCacher


def test_record_sets_hash(configured_test_manager):
    """A record that is passed an argset should appropriately set the hash on it immediately."""
    record = Record(configured_test_manager, ExperimentParameters(name="testing"))
    assert record.params.hash is not None


def test_record_doesnot_set_hash_for_none_args(configured_test_manager):
    """A record that is passed an argset should appropriately set the hash on it immediately."""
    record = Record(configured_test_manager, None)
    assert record.params is None


def test_record_stores_hash_when_not_dry(configured_test_manager):
    """A record that is passed an argset with a non-dry-mode manager should store the args in the
    param_registry.json."""
    record = Record(  # noqa: F841
        configured_test_manager, ExperimentParameters(name="testing")
    )

    reg_path = os.path.join(
        configured_test_manager.manager_cache_path, "params_registry.json"
    )
    with open(reg_path) as infile:
        reg = json.load(infile)

    keys = list(reg.keys())
    assert len(keys) == 1
    assert reg[keys[0]]["name"] == "testing"


def test_record_doesnot_store_hash_when_dry(configured_test_manager):
    """A record that is passed an argset with a dry-mode manager should not store the args in
    the param_registry.json."""
    configured_test_manager.dry = True
    record = Record(  # noqa: F841
        configured_test_manager, ExperimentParameters(name="testing")
    )

    reg_path = os.path.join(
        configured_test_manager.manager_cache_path, "params_registry.json"
    )
    assert not os.path.exists(reg_path)


def test_record_doesnot_store_hash_when_parallelmode(configured_test_manager):
    """A record that is passed an argset with a parallel-mode manager should not store the
    args in the param_registry.json."""
    configured_test_manager.parallel_mode = True
    record = Record(  # noqa: F841
        configured_test_manager, ExperimentParameters(name="testing")
    )

    reg_path = os.path.join(
        configured_test_manager.manager_cache_path, "params_registry.json"
    )
    assert not os.path.exists(reg_path)


def test_record_doesnot_store_combo_hash_when_parallel_mode(configured_test_manager):
    """In a parallel-mode run, the record should not store the combo hash of an
    aggregate stage in the param_registry.json."""
    configured_test_manager.parallel_mode = True

    @aggregate(None, ["testing"])
    def agg_stage(record, records):
        return "test"

    r0 = Record(configured_test_manager, None)
    r1 = Record(configured_test_manager, ExperimentParameters(name="test"))
    r0 = agg_stage(r0, [r1])

    reg_path = os.path.join(
        configured_test_manager.manager_cache_path, "params_registry.json"
    )
    assert not os.path.exists(reg_path)


# TODO: test_record_gets_combo_hash_for_aggregate (check that is_aggregate sets)
def test_record_gets_combo_hash_for_aggregate(configured_test_manager):
    @aggregate(None, ["testing"])
    def agg_stage(record, records):
        return "test"

    r0 = Record(configured_test_manager, None)
    r1 = Record(configured_test_manager, ExperimentParameters(name="test"))
    r0 = agg_stage(r0, [r1])
    combo_hash = hashing.add_params_combo_hash(
        r0, [r1], "", False
    )  # TODO: what about when None passed in? Empty array?

    assert r0.is_aggregate
    assert r0.combo_hash == combo_hash


# TODO: test_record_with_aggregate_doesnot_change_args_hash (do a normal stage with those args and an agg stage with those args)


def test_record_make_copy_retains_state(configured_test_manager):
    """When a record is copied, the resulting record should have the same values in state
    as the source record."""

    @stage([], outputs=["test"])
    def output_stage(record):
        return "hello world"

    r0 = Record(
        configured_test_manager, ExperimentParameters(name="test1")
    )  # TODO: include args
    r0 = output_stage(r0)

    r1 = r0.make_copy(ExperimentParameters(name="test2"))
    assert "test" in r1.state
    assert r1.state["test"] == "hello world"
    assert r1.params.name == "test2"


def test_record_make_copy_adds_record_to_manager(configured_test_manager):
    """Making a record copy with default values should add one (and only one) record
    to the manager."""

    @stage([], outputs=["test"])
    def output_stage(record):
        return "hello world"

    r0 = Record(configured_test_manager, ExperimentParameters(name="test0"))
    r0 = output_stage(r0)
    assert len(configured_test_manager.records) == 1

    r0.make_copy(ExperimentParameters(name="test1"))
    assert len(configured_test_manager.records) == 2


def test_record_make_copy_doesnot_add_record_to_manager_when_specified(
    configured_test_manager,
):
    """Making a record copy with default values should add one (and only one) record
    to the manager."""

    @stage([], outputs=["test"])
    def output_stage(record):
        return "hello world"

    r0 = Record(configured_test_manager, ExperimentParameters(name="test0"))
    r0 = output_stage(r0)
    assert len(configured_test_manager.records) == 1

    r0.make_copy(ExperimentParameters(name="test1"), add_to_manager=False)
    assert len(configured_test_manager.records) == 1


def test_record_has_access_to_cachers_in_stage(configured_test_manager):
    """Inside a stage, record.stage_cachers should reflect the instantiated
    cachers that will be used to save associated outputs."""

    @stage(outputs=["test1", "test2"], cachers=[JsonCacher, JsonCacher(prefix="wat")])
    def do_things(record):
        assert len(record.stage_cachers) == 2
        path1 = record.stage_cachers[0].get_path()
        path2 = record.stage_cachers[1].get_path()

        return path1, path2

    r0 = Record(configured_test_manager, ExperimentParameters(name="test0"))
    r0 = do_things(r0)
    assert r0.state["test1"] == "test/examples/data/cache/test_0_do_things_test1.json"
    assert r0.state["test2"] == "test/examples/data/cache/wat_0_do_things_test2.json"



"""Testing pathing functions on the manager."""

import os

from pytest_mock import mocker  # noqa: F401 -- flake8 doesn't see it's used as fixture

from curifactory import ExperimentParameters, Record, aggregate, hashing, stage
from curifactory.caching import JsonCacher, Lazy, PickleCacher

# -----------------------------------------
# get_artifact_path unit tests
# -----------------------------------------


def test_get_path_basic(sample_args, configured_test_manager):
    """Calling get_artifact_path with a basic set of args should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    path = configured_test_manager.get_artifact_path("test_output", record)
    assert path == "test/examples/data/cache/test_sample_hash__test_output"


def test_get_path_basic_w_stagename(sample_args, configured_test_manager):
    """Calling get_artifact_path with a basic set of args and a valid stage name should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    configured_test_manager.current_stage_name = "somestage"
    path = configured_test_manager.get_artifact_path("test_output", record)
    assert path == "test/examples/data/cache/test_sample_hash_somestage_test_output"


def test_get_path_basic_w_custom_stagename(sample_args, configured_test_manager):
    """Calling get_artifact_path with a specific stage name should override the current stage name in the
    returned path."""
    record = Record(configured_test_manager, sample_args)
    configured_test_manager.current_stage_name = "somestage"
    path = configured_test_manager.get_artifact_path(
        "test_output", record, stage_name="someotherstage"
    )
    assert (
        path == "test/examples/data/cache/test_sample_hash_someotherstage_test_output"
    )


def test_get_path_subdir(sample_args, configured_test_manager):
    """Calling get_artifact_path with a basic set of args and a subdir should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    path = configured_test_manager.get_artifact_path(
        "test_output", record, subdir="WHAT"
    )
    assert path == "test/examples/data/cache/WHAT/test_sample_hash__test_output"


def test_get_path_subdirs(sample_args, configured_test_manager):
    """Calling get_artifact_path with a basic set of args and multiple subdirs should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    path = configured_test_manager.get_artifact_path(
        "test_output", record, subdir="WHAT/somethingelse"
    )
    assert (
        path
        == "test/examples/data/cache/WHAT/somethingelse/test_sample_hash__test_output"
    )


def test_get_path_prefix(sample_args, configured_test_manager):
    """Calling get_artifact_path with a basic set of args and a prefix should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    path = configured_test_manager.get_artifact_path(
        "test_output", record, prefix="special_data_proc"
    )
    assert path == "test/examples/data/cache/special_data_proc_sample_hash__test_output"


def test_get_path_store_full(sample_args, configured_test_manager):
    """Calling get_artifact_path with store-full should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    configured_test_manager.store_full = True
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_artifact_path("test_output", record, store=True)
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/artifacts/test_sample_hash__test_output"
    )


def test_get_path_custom_name(sample_args, configured_test_manager):
    """Calling get_artifact_path when a custom name is in use should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    configured_test_manager.prefix = "some_custom_name"
    path = configured_test_manager.get_artifact_path("test_output", record)
    assert path == "test/examples/data/cache/some_custom_name_sample_hash__test_output"


def test_get_path_custom_name_and_store_full(sample_args, configured_test_manager):
    """Calling get_artifact_path when a custom name is in use and storefull is called should return the expected path."""
    record = Record(configured_test_manager, sample_args)
    configured_test_manager.store_full = True
    configured_test_manager.prefix = "some_custom_name"
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_artifact_path("test_output", record, store=True)
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/artifacts/some_custom_name_sample_hash__test_output"
    )


def test_get_path_no_args(configured_test_manager):
    """Calling get_artifact_path with None args will result in None in the output name."""
    record = Record(configured_test_manager, None)
    path = configured_test_manager.get_artifact_path("test_output", record)
    assert path == f"test/examples/data/cache/test_{record.combo_hash}__test_output"


# -----------------------------------------
# get_run_output_path unit tests
# -----------------------------------------


def test_get_run_output_path(configured_test_manager):
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_run_output_path()
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}"
    )


def test_get_run_output_path_w_obj_name(configured_test_manager):
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_run_output_path("testing_file.txt")
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/testing_file.txt"
    )
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}"
    )
    assert not os.path.isdir(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/testing_file.txt"
    )


def test_get_run_output_path_w_subdirs_obj_name(configured_test_manager):
    ts = configured_test_manager.get_str_timestamp()
    path = configured_test_manager.get_run_output_path("subs/testing_file.txt")
    assert (
        path
        == f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/subs/testing_file.txt"
    )
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/subs"
    )


# -----------------------------------------
# misc
# -----------------------------------------


def test_write_run_env_output(configured_test_manager):
    """Storing a manager run should output the four expected environment info textfiles into the store-full folder."""
    configured_test_manager.store()
    configured_test_manager.write_run_env_output()
    ts = configured_test_manager.get_str_timestamp()
    # NOTE: the run number is 1 because when we store it we actually populate that information (there's just a default value of 0)
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/requirements.txt"
    )
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/environment_meta.txt"
    )
    assert os.path.exists(
        f"test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/run_info.json"
    )


def test_run_line_sanitization_normal(configured_test_manager):
    """Ensure that a normal store-full run-line will result in a correct reproduction line in the run info."""
    configured_test_manager.run_line = "experiment test -p params1 --store-full"
    configured_test_manager.store_full = True
    configured_test_manager.store()
    ts = configured_test_manager.get_str_timestamp()
    assert (
        configured_test_manager.run_info["reproduce"]
        == f"experiment test -p params1 --cache test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/artifacts --dry-cache"
    )


def test_run_line_sanitization_order(configured_test_manager):
    """The placement of the --store-full flag in the initial run line should not impact the reproduction line."""
    configured_test_manager.run_line = (
        "experiment test -p params1 --store-full --parallel 4"
    )
    configured_test_manager.store_full = True
    configured_test_manager.store()
    ts = configured_test_manager.get_str_timestamp()
    assert (
        configured_test_manager.run_info["reproduce"]
        == f"experiment test -p params1 --parallel 4 --cache test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/artifacts --dry-cache"
    )


def test_run_line_sanitization_overwrite(configured_test_manager):
    """The reproduction line should not include overwrite."""
    configured_test_manager.run_line = (
        "experiment test -p params1 --store-full --parallel 4 --overwrite"
    )
    configured_test_manager.store_full = True
    configured_test_manager.store()
    ts = configured_test_manager.get_str_timestamp()
    assert (
        configured_test_manager.run_info["reproduce"]
        == f"experiment test -p params1 --parallel 4 --cache test/examples/data/runs/test_{configured_test_manager.experiment_run_number}_{ts}/artifacts --dry-cache"
    )


def test_cache_aware_dict_resolve(configured_test_manager):
    """Ensure when resolve is on, lazy objects in a record's state auto load the thing."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    output_stage(record)
    assert record.state["tester"] == "hello world"


def test_cache_aware_dict_no_resolve(configured_test_manager):
    """Ensure when resolve is off, lazy objects in a record's state remain lazy."""

    @stage([], [Lazy("tester")], cachers=[PickleCacher])
    def output_stage(record):
        return "hello world"

    record = Record(configured_test_manager, None)
    output_stage(record)
    record.state.resolve = False
    assert type(record.state["tester"]) == Lazy
    assert record.state["tester"].name == "tester"
    assert type(record.state["tester"].cacher) == PickleCacher


# -----------------------------------------
# hash-setting tests
# -----------------------------------------


# TODO: (05/09/2022) with and without an agg of None args
def test_aggregate_stage_record_uses_combo_hash(configured_test_manager):
    """A combo hash for an aggregate should be stored on the record."""

    @stage(None, ["normal_hash"])
    def normal_stage(record):
        return record.params.hash

    @aggregate(None, ["agg_hash"])
    def agg_stage(record, records):
        return record.params.hash

    r0 = Record(configured_test_manager, ExperimentParameters(name="test"))
    r0 = normal_stage(r0)

    # r1 = Record(configured_test_manager, None)
    r1 = Record(configured_test_manager, ExperimentParameters(name="test"))
    r1 = agg_stage(r1, [r0])

    assert r1.combo_hash is not None
    assert r1.combo_hash != r0.params.hash
    # TODO: move this to test_record


def test_stage_hash_after_aggregate_with_no_args(configured_test_manager):
    """Outputs from a normal stage after an aggregate stage with None arguments should be cached under the combo hash."""

    @aggregate(None, ["testing"], [JsonCacher])
    def agg_stage(record, records):
        return "test"

    @stage(["testing"], ["testing2"], [JsonCacher])
    def normal_stage(record, testing):
        return "test2"

    r0 = Record(configured_test_manager, None)
    r1 = Record(configured_test_manager, ExperimentParameters(name="test"))

    r0 = normal_stage(agg_stage(r0, [r1]))

    combo_hash = hashing.add_params_combo_hash(r0, [r1], "", False)
    output_path_agg = os.path.join(
        configured_test_manager.cache_path, f"test_{combo_hash}_agg_stage_testing.json"
    )
    output_path_normal = os.path.join(
        configured_test_manager.cache_path,
        f"test_{combo_hash}_normal_stage_testing2.json",
    )

    assert os.path.exists(output_path_agg)
    assert os.path.exists(output_path_normal)
    # TODO: move this to test_record?



import argparse
import sys

sys.path.append("test/examples")

from experiments import basic
from params.basic import Args
from stages.basic_stages import get_data, sum_data

from curifactory.experiment import main
from curifactory.manager import ArtifactManager
from curifactory.procedure import Procedure
from curifactory.utils import get_configuration


def test_import():
    """Importing the library should not throw errors!"""
    import curifactory  # noqa: F401 -- we're just trying to make sure it doesn't break


def test_ensure_config(configuration):
    """This is just to make sure the autouse fixture is doing what I expect."""
    assert get_configuration() == configuration


def test_experiment():
    manager = ArtifactManager(dry=True)
    basic.run(
        [
            Args(name="test1", starting_data=[1, 2, 3, 4]),
            Args(name="test2", starting_data=[4, 5, 6]),
        ],
        manager,
    )
    assert manager.records[0].state["sum"] == 10
    assert manager.records[1].state["sum"] == 15


def test_lone_procedure(configuration):
    proc = Procedure([get_data, sum_data], manager=ArtifactManager(dry=True))
    record = proc.run(Args(starting_data=[2, 3]))
    assert record.output == 5


def test_experiment_ls_output(mocker, capfd):
    mock = mocker.patch(  # noqa: F841
        "argparse.ArgumentParser.parse_args",
        return_value=argparse.Namespace(experiment_name="ls", parameters_name=""),
    )
    test = main()  # noqa: F841
    out, err = capfd.readouterr()
    assert (
        out
        == "EXPERIMENTS:\n\tbasic\n\tsubexp.example\n\nPARAMS:\n\tempty\n\tnonarrayargs\n\tparams1\n\tparams2\n\tsubparams.thing\n"
    )



"""Ensure the args hashing works as expected."""

from dataclasses import dataclass

import curifactory as cf
from curifactory.hashing import param_set_string_hash_representations


def test_args_subclass_hash_includes_all_sub_params():
    """The hash of a subclass of experiment args should include all of the subclass's
    parameters."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = None

    args1 = MyExperimentArgs(name="test", a=6, b=7)
    dry_hash_dict = args1.params_hash(dry=True)
    assert dry_hash_dict["a"] == ("repr(a)", "6")
    assert dry_hash_dict["b"] == ("repr(b)", "7")

    # make sure we correctly don't hash everything in the blacklist
    for should_skip in ["name", "hash", "overwrite", "hash_representations"]:
        assert dry_hash_dict[should_skip][0] == "SKIPPED: blacklist"

    # double check that different args with different params is in fact
    # a different hash
    args2 = MyExperimentArgs(a=5, b=6)
    assert args1.params_hash() != args2.params_hash()


def test_static_hashing_function_same_when_vals_diff():
    """Two instances of an args class where a value is different but the hashing mechanism
    is a function that returns the same value should both have the same hash."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = None

        hash_representations: dict = cf.set_hash_functions(a=lambda self, obj: 5)

    args1 = MyExperimentArgs()
    args2 = MyExperimentArgs(a=6)
    assert args1.params_hash() == args2.params_hash()


def test_none_hashing_function_same_when_vals_diff():
    """An argument in an args class with a hashing function set to none should not be
    taken into account in the output hash."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = 5

        hash_representations: dict = cf.set_hash_functions(a=None)

    args1 = MyExperimentArgs()
    args2 = MyExperimentArgs(a=6)
    assert args1.params_hash() == args2.params_hash()
    assert (
        args1.params_hash(dry=True)["a"][0]
        == "SKIPPED: set to None in hash_representations"
    )


def test_none_value_not_hashed():
    """An args class with a parameter set to None should not have that parameter included
    in the hash.

    This also demonstrates that you can add new arguments to an old class and old experiments
    will still run as long as the new args are defaulted to None.
    """

    @dataclass
    class MyExperimentArgs1(cf.ExperimentParameters):
        a: int = 0
        b: int = None

    @dataclass
    class MyExperimentArgs2(cf.ExperimentParameters):
        a: int = 0

    args1 = MyExperimentArgs1()
    args2 = MyExperimentArgs2()

    assert args1.params_hash() == args2.params_hash()
    assert args1.params_hash(dry=True)["b"][0] == "SKIPPED: value is None"


def test_new_param_none_hash_same_result():
    """Adding a new parameter to a parameterclass, but is added as None to hash representations should
    still hash to the same value.
    """

    @dataclass
    class MyExperimentArgs1(cf.ExperimentParameters):
        a: int = 0

    @dataclass
    class MyExperimentArgs2(cf.ExperimentParameters):
        a: int = 0
        b: int = 3

        hash_representations: dict = cf.set_hash_functions(b=None)

    args1 = MyExperimentArgs1()
    args2 = MyExperimentArgs2()

    assert args1.params_hash() == args2.params_hash()
    assert (
        args2.params_hash(dry=True)["b"][0]
        == "SKIPPED: set to None in hash_representations"
    )


def test_invalid_dataclassparam_should_not_change_hash():
    """Adding a parameter without a typehint (e.g. not a valid dataclass field) should not impact
    the hash."""

    @dataclass
    class MyExperimentArgs1(cf.ExperimentParameters):
        a: int = 0
        b = 3

        hash_representations: dict = cf.set_hash_functions(b=None)

    @dataclass
    class MyExperimentArgs2(cf.ExperimentParameters):
        a: int = 0
        b: int = 4

        hash_representations: dict = cf.set_hash_functions(b=None)

    args1 = MyExperimentArgs1()
    args2 = MyExperimentArgs2(b=7)

    assert args1.params_hash() == args2.params_hash()


def test_new_ignored_param_sub_dataclass():
    """Adding a new (none-ignored) parameter to a subdatclass should not change the overall hash."""

    @dataclass(unsafe_hash=True)
    class SubArgs1:
        a: int = 0

    @dataclass(unsafe_hash=True)
    class SubArgs2:
        a: int = 0
        b: int = 5

        hash_representations: dict = cf.set_hash_functions(b=None)

    @dataclass
    class Args1(cf.ExperimentParameters):
        sub: SubArgs1 = SubArgs1()

    @dataclass
    class Args2(cf.ExperimentParameters):
        sub: SubArgs2 = SubArgs2()

    a1 = Args1()
    a2 = Args2()

    assert a1.params_hash() == a2.params_hash()


def test_parameter_name_included_in_hash():
    """Parameter names should be involved in hashing - two args instances where parameters are the
    opposite of eachother should not hash to the same value."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = 1

    args1 = MyExperimentArgs(a=5, b=6)
    args2 = MyExperimentArgs(a=6, b=5)
    assert args1.params_hash() != args2.params_hash()


def test_custom_hashing_composed_dataclasses():
    """Composing multiple dataclasses into an experimentargs class should allow setting
    hashing functions on the other dataclasses."""

    @dataclass(unsafe_hash=True)
    class NormalDC:
        c: int = 5
        d: int = 6

        hash_representations: dict = cf.set_hash_functions(d=None)

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = 1
        others: NormalDC = NormalDC()

    args1 = MyExperimentArgs()
    args2 = MyExperimentArgs(others=NormalDC(d=7))
    assert args1.params_hash() == args2.params_hash()
    assert type(args1.params_hash(True)["others"][1]) == dict
    assert (
        args1.params_hash(True)["others"][1]["d"][0]
        == "SKIPPED: set to None in hash_representations"
    )


def test_composed_dataclasses_diff():
    """Composing multiple dataclasses into an experimentargs class should correctly
    change the hash if those sub arguments are different."""

    @dataclass(unsafe_hash=True)
    class NormalDC:
        c: int = 5
        d: int = 6

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = 1
        others: NormalDC = NormalDC()

    args1 = MyExperimentArgs()
    args2 = MyExperimentArgs(others=NormalDC(d=7))
    assert args1.params_hash() != args2.params_hash()


def test_set_hash_functions_with_kwargs():
    """calling set_hash_functions with kwargs should set a dictionary with the
    args as keys."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = None

        hash_representations: dict = cf.set_hash_functions(a=None, b=None)

    args = MyExperimentArgs()
    assert "a" in args.hash_representations
    assert "b" in args.hash_representations


def test_set_hash_functions_with_dict_arg():
    """calling set_hash_functions with a single dictionary should directly set
    the hash_representations."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = None

        hash_representations: dict = cf.set_hash_functions({"a": None, "b": None})

    args = MyExperimentArgs()
    assert "a" in args.hash_representations
    assert "b" in args.hash_representations


def test_set_hash_functions_with_dict_and_kwargs():
    """calling set_hash_functions with both a dictionary and kwargs should create
    a merged dictionary."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = None
        c: int = 5

        hash_representations: dict = cf.set_hash_functions(
            {"a": None, "b": None}, b="something else", c=None
        )

    args = MyExperimentArgs()
    assert "a" in args.hash_representations
    assert "b" in args.hash_representations
    assert "c" in args.hash_representations

    # kwargs should override what was in possitional arg
    assert args.hash_representations["b"] is not None

    """Subclassing an args class with hashing functions set and including additional
    hashing functions in the subclass should add/use the hashing functions of both."""

    """Subclassing an args class with hashing functions set and including hashing
    functions in the subclass for the same parameter name should 'override' parent
    where conflicting."""


def test_set_hash_functions_on_args_instance():
    """Setting the hashing functions directly on an instance of a parameter
    set should change that parameter's hash, but not of any other instance of
    those same parameters."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = None
        c: int = 5

    args0 = MyExperimentArgs()
    args1 = MyExperimentArgs()

    args0.hash_representations["c"] = None
    assert args0.params_hash() != args1.params_hash()


# TODO: (3/9/2023) I'm still unclear on if this should actually be the intended functionality
def test_hash_stays_same_after_param_change():
    """If you hash a parameter set, and then change a parameter the hash shouldn't change."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = None
        c: int = 5

    args = MyExperimentArgs()
    hash0 = args.params_hash()
    args.hash = hash0  # this emulates what run_experiment is doing.
    assert args.hash == hash0

    args.c = 3
    hash1 = args.params_hash()
    assert hash1 == hash0


def test_hash_changes_after_param_change_and_hash_set_to_none():
    """If you hash a parameter set, change a parameter, and set the .hash to `None`, the
    hash should recompute and then change."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = None
        c: int = 5

    args = MyExperimentArgs()
    hash0 = args.params_hash()
    args.hash = hash0  # this emulates what run_experiment is doing.
    assert args.hash == hash0

    args.c = 3
    args.hash = None
    hash1 = args.params_hash()
    assert hash1 != hash0


def test_none_hashing_function_includes_val_in_str_rep():
    """The string hash representation of an ignored parameter should still include the value
    in a sub IGNORED_PARAMS dictionary."""

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = 5

        hash_representations: dict = cf.set_hash_functions(a=None)

    args = MyExperimentArgs(a=6)
    rep = param_set_string_hash_representations(args)
    assert "a" in rep["IGNORED_PARAMS"]
    assert rep["IGNORED_PARAMS"]["a"] == "6"


def test_subdataclass_val_in_str_rep_correct():
    """The string hash rep of a dataclass with sub-dataclasses should correctly
    represent the sub dataclasses the same way."""

    @dataclass(unsafe_hash=True)
    class NormalDC:
        c: int = 5
        d: int = 6

        hash_representations: dict = cf.set_hash_functions(d=None)

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = 1
        others: NormalDC = NormalDC()

    args = MyExperimentArgs(others=NormalDC(d=7))
    rep = param_set_string_hash_representations(args)
    assert rep["others"]["c"] == "5"
    assert "d" in rep["others"]["IGNORED_PARAMS"]
    assert rep["others"]["IGNORED_PARAMS"]["d"] == "7"


def test_none_hash_subdataclass_val_in_str_rep_correct():
    """The string hash rep of a dataclass with sub-dataclasses should correctly
    represent the sub dataclasses the same way even if the sub dataclass is in the
    ignored parameters."""

    @dataclass(unsafe_hash=True)
    class NormalDC:
        c: int = 5
        d: int = 6

        hash_representations: dict = cf.set_hash_functions(d=None)

    @dataclass
    class MyExperimentArgs(cf.ExperimentParameters):
        a: int = 0
        b: int = 1
        others: NormalDC = NormalDC()
        hash_representations: dict = cf.set_hash_functions(others=None)

    args = MyExperimentArgs(others=NormalDC(d=7))
    rep = param_set_string_hash_representations(args)
    assert "others" in rep["IGNORED_PARAMS"]
    assert "d" in rep["IGNORED_PARAMS"]["others"]["IGNORED_PARAMS"]
    assert rep["IGNORED_PARAMS"]["others"]["IGNORED_PARAMS"]["d"] == "7"
